mod fd;
pub mod iso9660;

use crate::drivers::atapi::read_block;
use crate::fd::{FDt, FD_TABLE};
use crate::println;
use crate::serial_println;
use crate::utils::unserialize;

use super::FileSystem;
use fd::IsoFD;
use iso9660::{IsoDir, IsoPrimVolDesc};

use alloc::{boxed::Box, sync::Arc, string::String, vec::Vec};
use async_trait::async_trait;

pub struct IsoFS {}

#[async_trait(?Send)]
impl FileSystem for IsoFS {
    async fn open(&mut self, path: &str, flags: u32) -> Option<FDt> {
        // ISO is a read only file system
        if flags != crate::syscalls::io::O_RDONLY {
            return None;
        }

        let voldesc = get_prim_vol_desc().await;

        // Invalid ISO
        if voldesc.std_identifier != "CD001".as_bytes() {
            return None;
        }

        let root: IsoDir = voldesc.root_dir;
        let mut curr_entry_block: [u8; iso9660::ISO_BLOCK_SIZE] = read_block(root.data_blk.le).await;

        let mut curr_entry: &IsoDir = unserialize(curr_entry_block.as_ptr());

        let path_s: String = String::from(path);
        let path_it = path_s
            .split("/")
            .filter(|p| p != &"");

        for path_component in path_it {
            let mut found = false;
            while curr_entry.idf_len != 0 {
                serial_println!("{:?}", curr_entry.idf_len);
                serial_println!("{:?}", alloc::str::from_utf8(curr_entry.get_idf()).unwrap());

                // Found entry
                if curr_entry.matches(path_component) {
                    serial_println!("Found {}", path_component);
                    found = true;
                    curr_entry_block = read_block(curr_entry.data_blk.le).await;
                    curr_entry = unserialize(curr_entry_block.as_ptr());
                    break;
                }

                // Next entry
                curr_entry = curr_entry.next_entry();
            }
            if !found {
                return None;
            }
        }

        let fd = IsoFD::new();
        FD_TABLE.lock().await.register_fd(fd.clone());
        Some(fd.clone())
    }
}


pub async fn get_prim_vol_desc() -> IsoPrimVolDesc {
    let desc_block = read_block(iso9660::ISO_PRIM_VOLDESC_BLOCK).await;
    *unserialize::<IsoPrimVolDesc>(desc_block.as_ptr())
}
